#!venv/bin/python3
import pyAgrum as gum
import psycopg2 
import sys
import argparse

def is_var_indicator(v):
    return not is_var(v)

def is_var(v):
    return not v.lower().startswith('i')


parser = argparse.ArgumentParser()
parser.add_argument("BNFile", help="Bayesian network file")
parser.add_argument("DBSize", help="size of the database")
args = parser.parse_args()

# inputs
bn_file = args.BNFile
db_size = int(args.DBSize)
out_file = "mcc.csv"
table_name = "test"

# loading the BN file
bn = gum.loadBN(bn_file)
var_names = sorted(list(bn.names()))

# generating the CSV file
g=gum.BNDatabaseGenerator(bn)
g.drawSamples(db_size)
g.toCSV(out_file)

# PG connection
conn = psycopg2.connect(database="mcc", 
                        user='postgres', password='postgres',  
                        host='127.0.0.1', port='5432')
conn.autocommit = True
cursor = conn.cursor()

# deleting the existing table
deletion_sql = "DROP TABLE IF EXISTS {};".format(table_name)
cursor.execute(deletion_sql)

# creating the table to store the 
col_def = map(lambda name : "{} int NOT NULL".format(name), var_names)
creation_sql = "CREATE TABLE {}({});".format(table_name, ",".join(col_def))
cursor.execute(creation_sql) 

# loading the CSV data to the table
with open(out_file, "r") as file:
    next(file) # skip the header's line
    cursor.copy_from(file, table_name, sep=",", null="")

missing_selects = []
for v in var_names:
    if is_var(v) and v[1:] not in var_names:
        missing_selects.append("case when i{}=1 then NULL ELSE {} END".format(v,v))
    elif is_var(v):
        missing_selects.append(v)

missing_data_table = "{}_star".format(table_name)
deletion_sql = "DROP TABLE IF EXISTS {};".format(missing_data_table)
cursor.execute(deletion_sql)
missing_selects_text = ",".join(missing_selects)
missing_data_sql = "CREATE TABLE {} AS SELECT {}, md5(CAST(({}) AS text)) AS superblock FROM {}".format(missing_data_table, missing_selects_text, missing_selects_text, table_name)
print(missing_data_sql)
cursor.execute(missing_data_sql)

# superblock definitions
superblock_table = "{}_sb".format(table_name)
deletion_sql = "DROP TABLE IF EXISTS {};".format(superblock_table)
cursor.execute(deletion_sql)

vars = list(filter(is_var, var_names))
col_def = map(lambda name : "{} int NOT NULL".format(name), vars)
creation_sql = "CREATE TABLE {}(superblock text, {}, prob float NOT NULL);".format(superblock_table, ",".join(col_def))
cursor.execute(creation_sql) 


cursor.execute("SELECT DISTINCT * FROM {}".format(missing_data_table))
ie = gum.LazyPropagation(bn)

insert_sql = "INSERT INTO {}({}, superblock, prob) VALUES({},%s,%s)".format(superblock_table, ",".join(vars), ",".join(["%s" for _ in vars]))

def instantiation_to_list(inst, row, vars):
    l = []
    i = 0
    inst_vars = list(map(lambda v : v.name(), inst.variablesSequence()))
    for v in vars:
        if v in inst_vars:
            l.append(inst[v])
        else:
            l.append(row[i])
        i+=1
    return l

for row in cursor.fetchall():
    null_vars = []
    null_pos  = []
    i = 0
    for v in vars:
        if row[i] is None:
            null_vars.append(v)
            null_pos.append(i)
        else:
            ie.addEvidence(v, row[i])
        i+=1

    ie.addJointTarget(set(null_vars))
    if len(null_vars) > 0:
    #     if len(vars) > len(null_vars):
    #         print(set(null_vars))
    #         print(row)
    #         print(ie.jointTargets())
    #         print(ie.targets())
    #     else:
    #         print(vars)
        
        potentiel = ie.jointPosterior(set(null_vars)) if len(vars) > len(null_vars) else ie.evidenceJointImpact(vars,{})
        for i in potentiel.loopIn():
            inserted_row = instantiation_to_list(i, row, vars)
            inserted_row.append(row[-1])
            inserted_row.append(potentiel.get(i))
            cursor.execute(insert_sql, inserted_row)
    ie.eraseAllJointTargets()
    ie.eraseAllEvidence()

    
###### Imputation

# import numpy as np
# import pandas as pd
# from sklearn.impute import KNNImputer

# cursor.execute("SELECT * FROM {}".format(missing_data_table))
# tuples_list = cursor.fetchall()
# df = pd.DataFrame(tuples_list)

# imputer = KNNImputer(n_neighbors=10)
# impute = imputer.fit_transform(df)
# print(impute)
